;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "pet" "20260716.2214"
  "Executable and virtualenv tracker for python-mode."
  '((emacs "27.1")
    (f     "0.6.0")
    (map   "3.3.1")
    (seq   "2.24"))
  :url "https://github.com/wyuenho/emacs-pet/"
  :commit "c0d856782d74d205a9359ae2f5c623d8152d20d1"
  :revdesc "c0d856782d74"
  :keywords '("tools")
  :authors '(("Jimmy Yuen Ho Wong" . "wyuenho@gmail.com"))
  :maintainers '(("Jimmy Yuen Ho Wong" . "wyuenho@gmail.com")))
